"""Color format specifications for creating and interpreting representations of colors.

Often only describing a color system is not enough when dealing with libraries and frameworks
specifications. For example, both Kivy [1]_ and PyGame [2]_ use mostly RGB colors as inputs when
building applications. However, Kivy expects the colors to have a maximum value of 1, while PyGame
expects it to be 255. To bridge these expectations and allow the use of the same set of colors for
multiple projects across different frameworks, the :class:`ColorFormat` class was created.
This class can be used to build strictly defined formats which allows the creation and
interpretation of colors using :meth:`ColorFormat.new_color()` and :meth:`ColorFormat.format()`
respectively.

Interpreting color-like objects
-------------------------------

Most functions and classes of colorir can take as a parameter a type called
:const:`ColorLike`. This type stands for any common representation of color, such as those using
strings or tuples. The function/class is then able to interpret such a format using a color format
we provide or the default color format (see notes section at the end).

However, not all representation of a color can be understood by all :class:`ColorFormat` instances.
A tuple representation of red in RGB such as ``(255, 0, 0)``, for example, cannot be understood
by a color format such as ``ColorFormat(Hex)``. This is because there are many ways to represent
red as a tuple of three elements, such as ``(0, 1, 0.5)`` in HSL or as ``(1, 0, 0)`` in RGB with
a maximum value of 1 for each color spectrum, and a color format based on the Hex system is
incapable of diferentiating them. Because of this, for a color format to understand a tuple
(or list) representation of a color, it must itself be based on a tuple representation.

Other representations of color are unambiguous, and can safely be interpreted by any
:class:`ColorFormat`. A hex string of six elements such as ``"#ff0000"`` or ``"ff0000"`` is
understood universally, for example.

    .. list-table::
        :header-rows: 1

        * - Color representation
          - :class:`ColorFormat` that can understand it

        * - Tuple or list (e.g.: ``(255, 0, 0)``, ``(0, 1, 0.5)``)
          - Any tuple-based color format (e.g.: ``ColorFormat(sRGB)``,
            ``ColorFormat(HSL)``)

        * - Hex string with six elements (e.g.: ``"#ff0000"``)
          - Any

        * - Hex string with eight elements (e.g.: ``"#ff0000"``)
          - Hex string color format (e.g.: ``ColorFormat(Hex)``)

Built-in color formats
----------------------

Besides being able to specify your own color formats, a few pre-defined formats for popular
frameworks can be imported from this module, such as :const:`PYGAME_COLOR_FORMAT` and
:const:`WEB_COLOR_FORMAT`.

Examples:
    Import basic colors into a palette to work in a PyGame project:

    >>> from colorir import config, Palette, PYGAME_COLOR_FORMAT
    >>> config.DEFAULT_COLOR_FORMAT = PYGAME_COLOR_FORMAT
    >>> basic_colors = Palette.load("basic") # Loads colors as RGB values
    >>> # ... code

    Import the same colors but to work in a web project:

    >>> from colorir import WEB_COLOR_FORMAT
    >>> config.DEFAULT_COLOR_FORMAT = WEB_COLOR_FORMAT
    >>> basic_colors = Palette.load("basic") # Loads colors as hex codes
    >>> # ... code

    Create our own color format to work with HSL colors:

    >>> hsl_format = ColorFormat(color_class.HSL, max_h=100, max_sla=100, round_to=0)

    We can then create colors using the :meth:`ColorFormat.new_color()` and
    :meth:`ColorFormat.format()` methods:

    >>> hsl_format.new_color(30, 70, 70)
    HSL(30, 70, 70)

    >>> hsl_format.format((30, 70, 70))
    HSL(30, 70, 70)

Notes:
    To change the default color format to a custom or pre-built color format, take
    a look at :data:`config.DEFAULT_COLOR_FORMAT <colorir.config.DEFAULT_COLOR_FORMAT>`.

References:
    .. [1] Kivy: Cross-platform Python Framework for NUI Development at https://kivy.org/.
    .. [2] PyGame library at https://www.pygame.org/.
"""
from typing import Type, Union, NewType

from . import color_class

__all__ = [
    "ColorLike",
    "ColorFormat",
    "PYGAME_COLOR_FORMAT",
    "TKINTER_COLOR_FORMAT",
    "KIVY_COLOR_FORMAT",
    "WEB_COLOR_FORMAT",
    "MATPLOTLIB_COLOR_FORMAT"
]

ColorLike = NewType("ColorLike", Union["color_class.ColorBase", str, tuple, list])
"""Type constant that describes common representations of colors in python."""


class ColorFormat:
    """Class to create color format specifications.

    Args:
        color_sys: The color system that is the basis of this color format.
        format_kwargs: Keyword arguments that further specify the color format. These are specific
            to each color system and can be consulted in their respective documentations.

    Examples:
        >>> c_format = ColorFormat(color_class.sRGB, max_rgb=1)
        >>> c_format.new_color(1, 0, 0)
        sRGB(1, 0, 0)

        For more examples see the documentation of the :mod:`~colorir.color_format` module.
    """
    def __init__(self, color_sys: Type["color_class.ColorBase"], **format_kwargs):
        self.color_sys = color_sys
        self.format_params = format_kwargs

    def __getattr__(self, item):
        # __getattr__ is called only when super().__getattribute__ fails
        try:
            return self.format_params[item]
        except LookupError as e:
            raise AttributeError(f"AttributeError: '{self.__class__.__name__}' object has no "
                                 f"attribute '{item}'") from e

    def __call__(self, *args, **kwargs):
        return self.new_color(*args, **kwargs)

    def __repr__(self):
        param_strs = [f"{param}={val.__repr__()}" for param, val in self.format_params.items()]
        color_sys_str = f"color_sys={self.color_sys.__name__}"
        if param_strs:
            color_sys_str += ", "
        return f"{self.__class__.__name__}({color_sys_str}{', '.join(param_strs)})"

    def new_color(self, *args, **kwargs) -> "color_class.ColorBase":
        """Creates a new color by filling the remaining parameters of a color constructor with the
        provided arguments.

        Examples:
            >>> c_format = ColorFormat(color_class.sRGB,
            ...                        max_rgb=1,
            ...                        max_a=1,
            ...                        include_a=True)
            >>> c_format.new_color(1, 0, 0)
            sRGB(1, 0, 0, 1)
        """
        kwargs.update(self.format_params)
        return self.color_sys(*args, **kwargs)

    # Factory method to be called when reading the palette files or reconstructing colors
    def _from_rgba(self, rgba):
        return self.color_sys._from_rgba(rgba, **self.format_params)

    def format(self, color: ColorLike) -> "color_class.ColorBase":
        """Tries to format a color-like object into this color format.

        Because there are multiple implementations of tuple-based color systems, the `color`
        parameter is always allowed to be a subclass of :class:`~colorir.color_class.ColorBase` or
        a hex string with length == 6, but can only be a tuple if :attr:`ColorFormat.color_sys` is
        a tuple-based color system, such as :class:`~colorir.color_class.sRGB` for example.

        Examples:
            >>> rgb_format = ColorFormat(color_class.sRGB, round_to=0)
            >>> rgb_format.format("#ff0000")
            sRGB(255, 0, 0)
            >>> rgb_format.format((255, 0, 0))
            sRGB(255, 0, 0)
            >>> hex_format = ColorFormat(color_class.Hex)
            >>> hex_format.format("#ff0000")
            Hex('#ff0000')
            >>> hex_format.format((255, 0, 0)) # Can't understand how to parse this tuple
            Traceback (most recent call last):
              ...
            ValueError: tried to interpret a tuple-formatted color object with a Hex ColorFormat

        Args:
            color: The value of the color to be formatted. Can be an instance of any
                :mod:`~colorir.color_class` class or, alternatively, a color-like object that
                resembles the format of the color you want to format.
        """
        if isinstance(color, color_class.ColorBase):
            return self._from_rgba(color._rgba)
        elif isinstance(color, str):
            # Try to preserve input options (none implemented now but who knows)
            if self.color_sys == color_class.Hex:
                return self.new_color(color)
            # No alpha in the string, safe to interpret with Hex
            if len(color) < 8:
                # Fallback to Hex default args
                return self._from_rgba(color_class.Hex(color)._rgba)
            raise ValueError("tried to interpret a string-formatted color that contains an alpha"
                             "component with a non-Hex ColorFormat")
        if self.color_sys != color_class.Hex:
            # Assume that the color system is tuple-based
            return self.new_color(*color)
        else:
            raise ValueError("tried to interpret a tuple-formatted color object with a Hex "
                             "ColorFormat")


PYGAME_COLOR_FORMAT = ColorFormat(color_sys=color_class.sRGB, max_rgb=255, max_a=255, round_to=0)
"""Color format compatible with PyGame standards."""

KIVY_COLOR_FORMAT = ColorFormat(color_sys=color_class.sRGB, max_rgb=1, max_a=1, include_a=True)
"""Color format compatible with Kivy standards."""

MATPLOTLIB_COLOR_FORMAT = ColorFormat(color_sys=color_class.Hex, include_a=True, tail_a=True)
"""Color format compatible with Matplotlib."""

WEB_COLOR_FORMAT = ColorFormat(color_sys=color_class.Hex)
"""Color format compatible with HTML and CSS standards."""

TKINTER_COLOR_FORMAT = WEB_COLOR_FORMAT
"""Color format compatible with Tkinter standards."""



import pygame as pg
from colorir import Palette, PYGAME_COLOR_FORMAT
# Since pygame doesn't accept hex strings, we will change the color format of our palette to match
colors = Palette.load("turtles", color_format=PYGAME_COLOR_FORMAT)

pg.init()
clock = pg.time.Clock()
screen = pg.display.set_mode((720, 480))
while True:
    clock.tick()
    for event in pg.event.get():
        if event.type == pg.QUIT:
            quit()
    screen.fill(colors.dreamblue)  # Draw background
    # Draw legs
    pg.draw.rect(screen, colors.woodbrown, (300, 200, 75, 200))
    pg.draw.rect(screen, colors.woodbrown, (500, 200, 75, 200))
    # Draw shell
    pg.draw.ellipse(screen, colors.swampgreen, (275, 100, 325, 200))
    # Draw head
    pg.draw.circle(screen, colors.swampgreen, (275, 100), 70)
    pg.display.update()



import doctest
import unittest
from pathlib import Path
from random import randint
import re

from colorir import palette, config, Palette, StackPalette, HSV, find_palettes, delete_palette
config.DEFAULT_PALETTES_DIR = str(Path(__file__).resolve().parent / "test_palettes")
# Clear test palette directory
for file in Path(config.DEFAULT_PALETTES_DIR).glob("*"):
    # Keep original test files, remove others
    if re.search(r"test\d", file.name) is None:
        file.unlink()


class TestPalette(unittest.TestCase):
    def test_add_op(self):
        pal = Palette(c1="ffffff") + Palette(c2="000000")
        self.assertEqual(pal, Palette(c1="ffffff", c2="000000"))

    def test_save_load(self):
        colors = {f"c{i}": "%02x%02x%02x" % (randint(0, 255), randint(0, 255), randint(0, 255))
                  for i in range(250)}
        pal = Palette("test_sl", None, **colors)
        pal.save()
        pal2 = Palette.load("test_sl")
        self.assertEqual(pal, pal2)

    def test_load_warns(self):
        with self.assertWarns(Warning):
            Palette.load(palettes=["test1", "test2"], search_builtins=False)


class TestSwatchPalette(unittest.TestCase):
    def test_add_op(self):
        spal = StackPalette(None, None, "ffffff") + StackPalette(None, None, "000000")
        self.assertEqual(spal, StackPalette(None, None, "ffffff", "000000"))

    def test_save_load(self):
        colors = ["%02x%02x%02x" % (randint(0, 255), randint(0, 255), randint(0, 255))
                  for _ in range(250)]
        spal = StackPalette("test_sl", None, *colors)
        spal.save()
        spal2 = StackPalette.load("test_sl")
        self.assertEqual(spal, spal2)

    def test_complementary(self):
        red = "ff0000"
        spal = StackPalette.new_complementary(3, red)
        self.assertEqual(spal, StackPalette(None, None, red, "00ff00", "0000ff"))

    def test_analogous(self):
        red = HSV(0, 1, 1)
        # Center, clockwise and counter-clockwise generation
        c_spal = StackPalette.new_analogous(3, color=red)
        cw_spal = StackPalette.new_analogous(3, color=red, start=1)
        ccw_spal = StackPalette.new_analogous(3, color=red, start=-1)
        self.assertEqual(c_spal, StackPalette(None, None, HSV(330, 1, 1), red, HSV(30, 1, 1)))
        self.assertEqual(cw_spal, StackPalette(None, None, red, HSV(30, 1, 1), HSV(60, 1, 1)))
        self.assertEqual(ccw_spal, StackPalette(None, None, HSV(300, 1, 1), HSV(330, 1, 1), red))
        # Larger hue wheel sections
        wide_spal = StackPalette.new_analogous(3, sections=3, color=red)
        self.assertEqual(wide_spal, StackPalette(None, None, HSV(240, 1, 1), red, HSV(120, 1, 1)))


class TestOtherFunctions(unittest.TestCase):
    def test_find_palettes(self):
        all_pals = find_palettes(search_builtins=False)
        self.assertEqual(all_pals, ['test1', 'test2', 'test3'])
        pals = find_palettes(search_builtins=False, kind=Palette)
        self.assertEqual(pals, ['test1', 'test2'])
        spals = find_palettes(search_builtins=False, kind=StackPalette)
        self.assertEqual(spals, ['test3'])

    def test_delete_palette(self):
        n_pal = Palette("test_del", red="ff0000")
        n_pal.save()
        self.assertIn("test_del", find_palettes())
        delete_palette("test_del")
        self.assertNotIn("test_del", find_palettes())


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(palette))
    return tests


if __name__ == "__main__":
    unittest.main()


"""Gradients between colors.

Examples:
    Get purple inbetween red and blue:

    >>> grad = Grad(["ff0000", "0000ff"])
    >>> grad.perc(0.5)
    Hex('#be0090')

    Depending on your use-case, it may be useful to rather interpolate colors in a cilindrical
    color space, such as :class:`~colorir.color_class.HSV` or :class:`~colorir.color_class.HCLuv`.
    To do this, use :class:`PolarGrad`, which supports color lerping through polar coordinate
    color components:

    >>> p_grad = PolarGrad(["ff0000", "0000ff"])
    >>> p_grad.perc(0.5)
    Hex('#f000cc')

    Get 5 colors interspaced in the gradient:

    >>> grad.n_colors(5)
    [Hex('#ff0000'), Hex('#e00066'), Hex('#be0090'), Hex('#9400b9'), Hex('#0000ff')]
"""
import numpy as np
from typing import Iterable, Type

from . import config, utils
from .color_class import sRGB, CIELuv, ColorBase, HCLuv, ColorPolarBase
from .color_format import ColorLike

__all__ = [
    "Grad",
    "PolarGrad",
    "RGBGrad",
    "RGBLinearGrad"
]


class Grad:
    """Linear interpolation gradient.

    Args:
        colors: Iterable of colors that compose the gradient. There can be more than two colors in
            this iterable.
        color_format: Color format specifying how to output colors. Defaults to
            :const:`config.DEFAULT_COLOR_FORMAT <colorir.config.DEFAULT_COLOR_FORMAT>`.
        color_sys: Color system in which the colors will be interpolated.
    """

    def __init__(self,
                 colors: Iterable[ColorLike],
                 color_format=None,
                 color_sys: Type[ColorBase] = CIELuv):
        if color_format is None:
            color_format = config.DEFAULT_COLOR_FORMAT

        self.color_format = color_format
        self.color_sys = color_sys
        self.colors = [self.color_format.format(color) for color in colors]
        self._conv_colors = [color_sys._from_rgba(color._rgba, include_a=True)
                                  for color in self.colors]

    def perc(self, p: float):
        """Returns the color placed in a given percentage of the gradient.

        Args:
            p: Percentage of the gradient expressed in a range of 0-1 from which a color will be
                drawn.

        Examples:
            Get purple inbetween red and blue:

            >>> grad = Grad([sRGB(255, 0, 0), sRGB(0, 0, 255)])
            >>> grad.perc(0.5)
            Hex('#be0090')

            Get a very "reddish" purple:

            >>> grad.perc(0.2)
            Hex('#e6005c')
        """
        i = int(p * (len(self._conv_colors) - 1))
        new_color = self._linear_interp(
            self._conv_colors[i],
            self._conv_colors[min([i + 1, len(self._conv_colors) - 1])],
            p * (len(self._conv_colors) - 1) - i
        )
        # Sometimes this reinterpretation makes colors seem to not be linearly-interpolated
        # HCLuv(287, 99, 31) -> sRGB(131, 0, 185) -> HCLuv(286, 95, 35)
        # This is okay though
        return self.color_format._from_rgba(new_color._rgba)

    def n_colors(self, n: int, include_ends=True):
        """Return `n` interspaced colors from the gradient.

        Args:
            n: Number of colors to retrieve.
            include_ends: Stops the colors at the extreme of the gradient from being sampled.
                This allows sampling a small number of colors (such as two) without having it
                return the same colors that were used to create the gradient in the first place.
        """
        colors = []
        sub = -1 if include_ends else 1
        for i in range(n):
            p = (i + (not include_ends)) / (n + sub)
            colors.append(self.perc(p))
        return colors

    def to_cmap(self, name, N=256, gamma=1.0):
        """Converts this gradient into a matplotlib LinearSegmentedColormap.

        Args:
            name: Passed down to ListedColormap constructor.
            N: Passed down to ListedColormap constructor.
            gamma: Passed down to ListedColormap constructor.
        """
        from matplotlib.colors import LinearSegmentedColormap

        colors = [color.hex(include_a=True, tail_a=True) for color in self.n_colors(N)]
        return LinearSegmentedColormap.from_list(name=name, colors=colors, N=N, gamma=gamma)

    def _linear_interp(self, color1, color2, p: float):
        """Receives two colors in 'self.color_sys' format (expected to include_a) and returns a
        new color in the same format."""
        color1, color2 = np.array(color1), np.array(color2)
        return self.color_sys(
            *(color1 + (color2 - color1) * p)
        )


# TODO doc
class PolarGrad(Grad):
    def __init__(self,
                 colors,
                 color_format=None,
                 color_sys=HCLuv,
                 lerp=True):
        if not issubclass(color_sys, ColorPolarBase):
            raise ValueError("'color_sys' must be a subclass of 'ColorPolarBase'")

        super().__init__(colors=colors, color_format=color_format, color_sys=color_sys)
        self.lerp = lerp

    def _linear_interp(self, color1, color2, p: float):
        d = abs(color2.h - color1.h)
        if not self.lerp or d <= color1.max_h / 2:
            return super()._linear_interp(color1=color1, color2=color2, p=p)

        array1, array2 = np.array(color1), np.array(color2)
        if color1.h > color2.h:
            array1, array2 = array2, array1
            p = 1 - p
        array1[0] += color1.max_h
        new_color = array1 + (array2 - array1) * p
        new_color[0] %= color1.max_h

        return self.color_sys(*new_color)


class RGBGrad(Grad):
    """Linear interpolation gradient using the RGB color space.

    This class allows colors to be interpolated using linear RGB through the 'use_linear_rgb'
    parameter. Otherwise, ``RGBGrad(..., use_linear_rgb=False)`` is the same as
    ``Grad(..., color_sys=sRGB)``.

    Args:
        colors: Iterable of colors that compose the gradient. There can be more than two colors in
            this iterable.
        color_format: Color format specifying how to output colors. Defaults to
            :const:`config.DEFAULT_COLOR_FORMAT <colorir.config.DEFAULT_COLOR_FORMAT>`.
        use_linear_rgb: Whether to use linear RGB rather than sRGB to create the gradient.

    Notes:
        Often using linear RGB may result in a gradient that looks more natural to the human eye
        [#]_

    References:
        .. [#] Ayke van Laethem at https://aykevl.nl/2019/12/colors
    """

    def __init__(self, colors: Iterable[ColorLike], color_format=None, use_linear_rgb=True):
        super().__init__(colors=colors, color_format=color_format, color_sys=sRGB)
        self.use_linear_rgb = use_linear_rgb

    def _linear_interp(self, color1, color2, p: float):
        if self.use_linear_rgb:
            rgba_1 = utils._to_linear_rgb(color1._rgba)
            rgba_2 = utils._to_linear_rgb(color2._rgba)
        else:
            rgba_1 = color1._rgba
            rgba_2 = color2._rgba

        new_rgba = tuple(rgba_1[i] + (rgba_2[i] - rgba_1[i]) * p for i in range(4))
        if self.use_linear_rgb:
            new_rgba = utils._to_srgb(new_rgba)
        return sRGB._from_rgba(new_rgba)


# Alias
RGBLinearGrad = RGBGrad